/* Copyright (C) 2013 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.example.games.tq;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Date;

import android.os.Bundle;
import android.util.Log;
import android.view.View;
import android.widget.TextView;

import com.google.android.gms.appstate.AppStateClient;
import com.google.android.gms.appstate.OnStateDeletedListener;
import com.google.android.gms.appstate.OnStateLoadedListener;
import com.google.android.gms.games.GamesClient;
import com.google.android.gms.games.achievement.Achievement;
import com.google.android.gms.games.achievement.AchievementBuffer;
import com.google.android.gms.games.achievement.OnAchievementsLoadedListener;
import com.google.example.games.basegameutils.BaseGameActivity;

/**
 * Trivial quest. A sample game that sets up the Google Play game services
 * API and allows the user to click a button to win (yes, incredibly epic).
 * Even though winning the game is fun, the purpose of this sample is to
 * illustrate the simplest possible game that uses the API.
 *
 * @author Bruno Oliveira (Google)
 *
 */
public class MainActivity extends BaseGameActivity implements View.OnClickListener,
        OnStateLoadedListener, OnStateDeletedListener, OnAchievementsLoadedListener {
    private static boolean DEBUG_ENABLED = true;
    private static final String TAG = "TrivialQuest";
    private static final int STATE_LAST_UPDATE = 0;
    private TextView textview;

    public MainActivity() {
        // request GamesClient and AppStateClient
        super(BaseGameActivity.CLIENT_GAMES | BaseGameActivity.CLIENT_APPSTATE);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        enableDebugLog(DEBUG_ENABLED, TAG);
        super.onCreate(savedInstanceState);

        setContentView(R.layout.activity_main);
        findViewById(R.id.button_sign_in).setOnClickListener(this);
        findViewById(R.id.button_sign_out).setOnClickListener(this);
        findViewById(R.id.button_win).setOnClickListener(this);
        findViewById(R.id.button_save_to_cloud).setOnClickListener(this);
        findViewById(R.id.button_load_from_cloud).setOnClickListener(this);
        findViewById(R.id.button_delete_from_cloud).setOnClickListener(this);
        findViewById(R.id.button_load_achievements).setOnClickListener(this);

        textview = (TextView) findViewById(R.id.textview_state);
    }

    // Shows the "sign in" bar (explanation and button).
    private void showSignInBar() {
        findViewById(R.id.sign_in_bar).setVisibility(View.VISIBLE);
        findViewById(R.id.sign_out_bar).setVisibility(View.GONE);
    }

    // Shows the "sign out" bar (explanation and button).
    private void showSignOutBar() {
        findViewById(R.id.sign_in_bar).setVisibility(View.GONE);
        findViewById(R.id.sign_out_bar).setVisibility(View.VISIBLE);
    }

    /**
     * Called to notify us that sign in failed. Notice that a failure in sign in is not
     * necessarily due to an error; it might be that the user never signed in, so our
     * attempt to automatically sign in fails because the user has not gone through
     * the authorization flow. So our reaction to sign in failure is to show the sign in
     * button. When the user clicks that button, the sign in process will start/resume.
     */
    @Override
    public void onSignInFailed() {
        // Sign-in has failed. So show the user the sign-in button
        // so they can click the "Sign-in" button.
        showSignInBar();
    }

    /**
     * Called to notify us that sign in succeeded. We react by loading the loot from the
     * cloud and updating the UI to show a sign-out button.
     */
    @Override
    public void onSignInSucceeded() {
        // Sign-in worked!
        showSignOutBar();
    }

    @Override
    public void onClick(View view) {
        switch (view.getId()) {
        case R.id.button_sign_in:
            // Check to see the developer who's running this sample code read the instructions :-)
            // NOTE: this check is here only because this is a sample! Don't include this
            // check in your actual production app.
            if (!verifyPlaceholderIdsReplaced()) {
                showAlert("Error", "Sample not correctly set up. See README!");
                break;
            }

            // start the sign-in flow
            beginUserInitiatedSignIn();
            break;
        case R.id.button_sign_out:
            // sign out.
            signOut();
            showSignInBar();
            break;
        case R.id.button_win:
            // win!
            showAlert(getString(R.string.victory), getString(R.string.you_won));
            if (getGamesClient().isConnected()) {
                // unlock the "Trivial Victory" achievement.
                getGamesClient().unlockAchievement(
                        getString(R.string.trivial_victory_achievement_id));
            }
            break;
        case R.id.button_save_to_cloud:
            // save to cloud
            long val = System.currentTimeMillis();
            getAppStateClient().updateState(STATE_LAST_UPDATE, toBytes(val));
            displayMessage("Save data: " + val);
            break;
        case R.id.button_load_from_cloud:
            // load from cloud
            getAppStateClient().loadState(this, STATE_LAST_UPDATE);
            break;
        case R.id.button_delete_from_cloud:
            // delete from cloud
            getAppStateClient().deleteState(this, STATE_LAST_UPDATE);
            break;
        case R.id.button_load_achievements:
            // load achievements
            getGamesClient().loadAchievements(this, true);
            break;
        }
    }

    /**
     * Checks that the developer (that's you!) read the instructions.
     *
     * IMPORTANT: a method like this SHOULD NOT EXIST in your production app!
     * It merely exists here to check that anyone running THIS PARTICULAR SAMPLE
     * did what they were supposed to in order for the sample to work.
     */
    boolean verifyPlaceholderIdsReplaced() {
        final boolean CHECK_PKGNAME = true; // set to false to disable check (not recommended!)

        // Did the developer forget to change the package name?
        if (CHECK_PKGNAME && getPackageName().startsWith("com.google.example.")) return false;

        // Did the developer forget to replace a placeholder ID?
        int res_ids[] = new int[] {
            R.string.app_id, R.string.trivial_victory_achievement_id
        };

        for (int i : res_ids) {
            if (getString(i).equalsIgnoreCase("ReplaceMe")) return false;
        }
        return true;
    }

    private long toLong(byte[] data) {
        if (data == null) {
            return -1;
        }
        ByteArrayInputStream bais = new ByteArrayInputStream(data);
        DataInputStream dis = new DataInputStream(bais);
        try {
            return dis.readLong();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private byte[] toBytes(long val) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        DataOutputStream dos = new DataOutputStream(baos);
        try {
            dos.writeLong(val);
            return baos.toByteArray();
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private void displayMessage(String msg) {
        textview.setText(msg);
    }

    @Override
    public void onStateLoaded(int statusCode, int stateKey, byte[] data) {
        Log.d(TAG, "onStateLoaded");

        if (statusCode == AppStateClient.STATUS_OK) {
            // successfully loaded data
            displayMessage("Loaded: " + toLong(data));
        } else if (statusCode == AppStateClient.STATUS_NETWORK_ERROR_STALE_DATA) {
            // could not connect to get fresh data,
            // but loaded (possibly out-of-sync) cached data instead
            displayMessage("Stale data: " + toLong(data));
        } else {
            // handle error
            displayMessage("Data load error: " + statusCode);
        }
    }

    @Override
    public void onStateConflict(int stateKey, String resolvedVersion, byte[] localData, byte[] serverData) {
        Log.d(TAG, "onStateConflict");

        long localVal = toLong(localData);
        long serverVal = toLong(serverData);
        long resolvedVal = Math.max(localVal, serverVal);
        byte[] resolvedData = toBytes(resolvedVal);
        getAppStateClient().resolveState(this, STATE_LAST_UPDATE, resolvedVersion, resolvedData);
        displayMessage("Conflicted: " + resolvedVal);
    }

    @Override
    public void onStateDeleted(int statusCode, int stateKey) {
        Log.d(TAG, "onStateDeleted");
        if (statusCode == AppStateClient.STATUS_OK) {
            displayMessage("Deleted");
        } else {
            displayMessage("Delete error: " + statusCode);
        }
    }

    @Override
    public void onAchievementsLoaded(int statusCode, AchievementBuffer buffer) {
        if (statusCode == GamesClient.STATUS_OK) {
            int achievementsCount = buffer.getCount();
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < achievementsCount; i++) {
                Achievement achievement = buffer.get(i);
                sb.append("Name: " + achievement.getName() + "\n");
                sb.append("Player: " + achievement.getPlayer().getDisplayName() + "\n");
                sb.append("Date: " + new Date(achievement.getLastUpdatedTimestamp()));
                Log.d(TAG, achievement.toString());
            }
            displayMessage(new String(sb));
        } else {
            // handle error
            displayMessage("Achievement load error: " + statusCode);
        }
    }
}
